import argparse
import asyncio
import os
import sys

from loguru import logger

from alist_mikananirss import (
    AnimeRenamer,
    AppConfig,
    BotAssistant,
    ConfigManager,
    DownloadManager,
    NotificationSender,
    RegexFilter,
    RemapperManager,
    RssMonitor,
    SubscribeDatabase,
)
from alist_mikananirss.alist import Alist
from alist_mikananirss.bot import BotFactory, NotificationBot
from alist_mikananirss.extractor import Extractor, LLMExtractor, create_llm_provider


def init_logging(cfg: AppConfig):
    log_level = cfg.dev.log_level
    logger.remove()

    # 确保日志目录存在
    os.makedirs("log", exist_ok=True)

    # 使用loguru的动态日期格式化功能
    log_filename = "log/alist_mikanrss_{time:YYYY-MM-DD}.log"
    logger.add(
        log_filename, rotation="00:00", retention="7 days", level=log_level, mode="a"
    )
    logger.add(sys.stderr, level=log_level)


def init_proxies(cfg: AppConfig):
    proxies = cfg.common.proxies
    if not proxies:
        return
    if "http" in proxies:
        os.environ["HTTP_PROXY"] = proxies["http"]
    if "https" in proxies:
        os.environ["HTTPS_PROXY"] = proxies["https"]


def init_notification(cfg: AppConfig):
    notification_bots = []
    if not cfg.notification.enable:
        return

    for bot_cfg in cfg.notification.bots:
        bot_kwargs = bot_cfg.model_dump(exclude={"bot_type"})

        try:
            bot = BotFactory.create_bot(bot_cfg.bot_type, **bot_kwargs)
            notification_bots.append(NotificationBot(bot))
        except ValueError as e:
            logger.error(f"Failed to create notification bot: {e}")

    if not notification_bots:
        logger.warning(
            "Notification enabled but no valid bots were configured or created."
        )
        cfg.notification.enable = False
        return

    NotificationSender.initialize(notification_bots, cfg.notification.interval_time)


async def run():
    parser = argparse.ArgumentParser(description="Alist Mikanani RSS")
    parser.add_argument(
        "--config",
        default="config.yaml",
        help="Path to the configuration file",
    )

    args = parser.parse_args()

    cfg_manager = ConfigManager()
    cfg = cfg_manager.load_config(args.config)
    # logger
    init_logging(cfg)

    logger.info("Loaded config Successfully")
    logger.info(f"Config: \n{cfg}")

    # proxy
    init_proxies(cfg)

    # database
    db = await SubscribeDatabase.create()

    # alist
    alist_client = Alist(cfg.alist.base_url, cfg.alist.token, cfg.alist.downloader)
    alist_ver = await alist_client.get_alist_ver()
    if alist_ver < "3.42.0":
        raise ValueError(f"Unsupported Alist version: {alist_ver}")

    # download manager
    DownloadManager.initialize(
        alist_client=alist_client,
        base_download_path=cfg.alist.download_path,
        use_renamer=cfg.rename.enable,
        need_notification=cfg.notification.enable,
        db=db,
    )

    # extractor
    if cfg.rename.enable:
        extractor_cfg = cfg.rename.extractor
        provider_kwargs = extractor_cfg.model_dump(
            exclude={"extractor_type", "output_type"}
        )
        try:
            llm_provider = create_llm_provider(
                extractor_cfg.extractor_type, **provider_kwargs
            )
            extractor = LLMExtractor(llm_provider, extractor_cfg.output_type)
        except ValueError as e:
            logger.error(f"Failed to create LLM provider: {e}")
            raise
        Extractor.initialize(extractor)

        AnimeRenamer.initialize(alist_client, cfg.rename.rename_format)

    # remapper
    if cfg.rename.remap.enable:
        cfg_path = cfg.rename.remap.cfg_path
        RemapperManager.load_remappers_from_cfg(cfg_path)

    # rss monitor
    regex_filter = RegexFilter()
    filters_name = cfg.mikan.filters
    regex_pattern = cfg.mikan.regex_pattern
    regex_filter.update_regex(regex_pattern)
    for name in filters_name:
        regex_filter.add_pattern(name)

    subscribe_url = cfg.mikan.subscribe_url
    rss_monitor = RssMonitor(
        subscribe_urls=subscribe_url,
        db=db,
        filter=regex_filter,
        use_extractor=cfg.rename.enable,
    )
    rss_monitor.set_interval_time(cfg.common.interval_time)

    tasks = []
    tasks.append(rss_monitor.run())
    # notification
    if cfg.notification.enable:
        init_notification(cfg)
        tasks.append(NotificationSender.run())

    # Initialize bot assistant
    if cfg.bot_assistant.enable:
        # Only telegram bot is supported now
        if cfg.bot_assistant.bots[0].bot_type == "telegram":
            bot_assistant = BotAssistant(cfg.bot_assistant.bots[0].token, rss_monitor)
            tasks.append(bot_assistant.run())

    try:
        await asyncio.gather(*tasks)
    finally:
        # cleanup after program exit
        await db.close()
        await alist_client.close()
        if cfg.bot_assistant.enable:
            await bot_assistant.stop()


def main():
    asyncio.run(run())



from unittest.mock import patch

import feedparser
import pytest

from alist_mikananirss.extractor import (
    AnimeNameExtractResult,
    ResourceTitleExtractResult,
    VideoQuality,
)
from alist_mikananirss.websites.mikan import Mikan, MikanHomePageInfo
from alist_mikananirss.websites.models import FeedEntry, LanguageType, ResourceInfo


@pytest.fixture
def mikan():
    return Mikan("https://mikanani.me/RSS/Bangumi?bangumiId=3519&subgroupid=382")


@pytest.fixture
def mock_rss_data():
    return """<rss version="2.0">
<channel>
<title>Mikan Project - GIRLS BAND CRY</title>
<link>https://mikanani.me/RSS/Bangumi?bangumiId=3298&subgroupid=382</link>
<description>Mikan Project - GIRLS BAND CRY</description>
<item>
<guid isPermaLink="false">【喵萌Production】★04月新番★[GIRLS BAND CRY][01-13][1080p][繁日双语][招募翻译时轴]</guid>
<link>https://mikanani.me/Home/Episode/a19d5da34e2ec205bddd9c6935ab579ff37da7d7</link>
<title>【喵萌Production】★04月新番★[GIRLS BAND CRY][01-13][1080p][繁日双语][招募翻译时轴]</title>
<description>【喵萌Production】★04月新番★[GIRLS BAND CRY][01-13][1080p][繁日双语][招募翻译时轴][8.8GB]</description>
<torrent xmlns="https://mikanani.me/0.1/">
<link>https://mikanani.me/Home/Episode/a19d5da34e2ec205bddd9c6935ab579ff37da7d7</link>
<contentLength>9448928256</contentLength>
<pubDate>2024-07-17T19:09:00</pubDate>
</torrent>
<enclosure type="application/x-bittorrent" length="9448928256" url="https://mikanani.me/Download/20240717/a19d5da34e2ec205bddd9c6935ab579ff37da7d7.torrent"/>
</item>
</channel>
</rss>"""


@pytest.mark.asyncio
async def test_get_feed_entries(mikan, mock_rss_data):
    with patch.object(
        mikan, "parse_feed", return_value=feedparser.parse(mock_rss_data)
    ):
        result = await mikan.get_feed_entries()

    assert isinstance(result, list)
    assert len(result) == 1
    entry = result.pop()
    assert isinstance(entry, FeedEntry)
    assert (
        entry.resource_title
        == "【喵萌Production】★04月新番★[GIRLS BAND CRY][01-13][1080p][繁日双语][招募翻译时轴]"
    )
    assert (
        entry.torrent_url
        == "https://mikanani.me/Download/20240717/a19d5da34e2ec205bddd9c6935ab579ff37da7d7.torrent"
    )
    assert entry.published_date == "2024-07-17T19:09:00"
    assert (
        entry.homepage_url
        == "https://mikanani.me/Home/Episode/a19d5da34e2ec205bddd9c6935ab579ff37da7d7"
    )


@pytest.mark.asyncio
async def test_get_feed_entries_real(mikan):
    # 测试真实的蜜柑RSS链接解析是否有报错
    await mikan.get_feed_entries()


@pytest.mark.asyncio
async def test_parse_homepage_error(mikan):
    # 对于蜜柑，强需求主页中的番剧名/字幕组信息，解析详情页时，如果出现异常，应该抛出异常
    mock_entry = FeedEntry(
        resource_title="【喵萌Production】★04月新番★[GIRLS BAND CRY][01-13][1080p][繁日双语][招募翻译时轴]",
        torrent_url="https://mikanani.me/Download/20240717/a19d5da34e2ec205bddd9c6935ab579ff37da7d7.torrent",
        published_date="2024-07-17T19:09:00",
        homepage_url="https://mikanani.me/Home/Episode/a19d5da34e2ec205bddd9c6935ab579ff37da7d7",
    )
    with patch.object(mikan, "parse_homepage", side_effect=Exception):
        with pytest.raises(Exception):
            await mikan.extract_resource_info(mock_entry, use_extractor=False)


@pytest.mark.asyncio
async def test_extract_resource_info(mikan):
    # 测试蜜柑是否使用主页提供的番剧名/字幕组信息
    mock_entry = FeedEntry(
        resource_title="【喵萌Production】★04月新番★[GIRLS BAND CRY][01][1080p][繁日双语][招募翻译时轴]",
        torrent_url="https://mikanani.me/Download/20240717/a19d5da34e2ec205bddd9c6935ab579ff37da7d7.torrent",
        published_date="2024-07-17T19:09:00",
        homepage_url="https://mikanani.me/Home/Episode/a19d5da34e2ec205bddd9c6935ab579ff37da7d7",
    )

    mock_homepage_info = MikanHomePageInfo(
        anime_name="GIRLS BAND CRY", fansub="喵萌奶茶屋"
    )

    mock_animename_extract_result = AnimeNameExtractResult(
        anime_name="GIRLS BAND CRY", season=1
    )

    mock_extract_result = ResourceTitleExtractResult(
        anime_name="tmdb_name",
        season=1,
        episode=1,
        quality=VideoQuality.p1080,
        languages=[LanguageType.SIMPLIFIED_CHINESE, LanguageType.JAPANESE],
        fansub="gpt_fansub",
        version=1,
    )

    with patch.object(mikan, "parse_homepage", return_value=mock_homepage_info):
        with patch(
            "alist_mikananirss.extractor.Extractor.analyse_resource_title",
            return_value=mock_extract_result,
        ):
            with patch(
                "alist_mikananirss.extractor.Extractor.analyse_anime_name",
                return_value=mock_animename_extract_result,
            ):
                result = await mikan.extract_resource_info(
                    mock_entry, use_extractor=True
                )

    assert isinstance(result, ResourceInfo)
    assert result.anime_name == mock_homepage_info.anime_name
    assert result.resource_title == mock_entry.resource_title
    assert result.torrent_url == mock_entry.torrent_url
    assert result.season == mock_extract_result.season
    assert result.episode == mock_extract_result.episode
    assert result.fansub == mock_homepage_info.fansub



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Type, TypeVar

T = TypeVar("T")


class LLMProvider(ABC):
    """Base class for LLM providers"""

    @abstractmethod
    async def parse_with_schema(
        self, messages: List[Dict[str, str]], response_format: Type[T]
    ) -> Optional[T]:
        """Parse response with a schema"""
        pass

    @abstractmethod
    async def parse_as_json(self, messages: List[Dict[str, str]]) -> Dict[str, Any]:
        """Parse response as a JSON object"""
        pass



import asyncio
from unittest.mock import AsyncMock

import pytest

from alist_mikananirss.bot import NotificationBot
from alist_mikananirss.core import NotificationSender
from alist_mikananirss.websites.models import ResourceInfo


@pytest.fixture(autouse=True)
def reset_notification_sender():
    NotificationSender.destroy_instance()


@pytest.mark.asyncio
async def test_initialization():
    mock_bot = AsyncMock(spec=NotificationBot)
    NotificationSender.initialize([mock_bot], interval=30)
    instance1 = NotificationSender()
    instance2 = NotificationSender()
    assert instance1 == instance2
    assert len(instance1.notification_bots) == 1
    assert instance1._interval == 30
    assert isinstance(instance1._queue, asyncio.Queue)


@pytest.mark.asyncio
async def test_set_notification_bots():
    NotificationSender.initialize([])
    mock_bot1 = AsyncMock(spec=NotificationBot)
    mock_bot2 = AsyncMock(spec=NotificationBot)

    NotificationSender.set_notification_bots([mock_bot1, mock_bot2])

    assert len(NotificationSender().notification_bots) == 2


@pytest.mark.asyncio
async def test_add_resource():
    NotificationSender.initialize([])
    resource = ResourceInfo(
        anime_name="test name",
        resource_title="Test",
        torrent_url="https://test.com",
        published_date="2023-01-01",
    )

    await NotificationSender.add_resource(resource)

    assert NotificationSender()._queue.qsize() == 1


@pytest.mark.asyncio
async def test_set_interval():
    NotificationSender.initialize([], interval=60)

    NotificationSender.set_interval(120)

    assert NotificationSender()._interval == 120


@pytest.mark.asyncio
async def test_run_method():
    mock_bot = AsyncMock(spec=NotificationBot)
    mock_bot.send_message.return_value = asyncio.Future()
    mock_bot.send_message.return_value.set_result(None)

    NotificationSender.initialize([mock_bot], interval=0.1)
    resource = ResourceInfo(
        anime_name="test name",
        resource_title="Test",
        torrent_url="https://test.com",
        published_date="2023-01-01",
    )
    await NotificationSender.add_resource(resource)
    task = asyncio.create_task(NotificationSender.run())
    await asyncio.sleep(0.2)

    mock_bot.send_message.assert_called_once()
    task.cancel()


@pytest.mark.asyncio
async def test_send_method_without_bots():
    NotificationSender.initialize([], interval=0.1)
    resource = ResourceInfo(
        anime_name="test name",
        resource_title="Test",
        torrent_url="https://test.com",
        published_date="2023-01-01",
    )
    await NotificationSender.add_resource(resource)
    task = asyncio.create_task(NotificationSender.run())
    await asyncio.sleep(0.2)
    task.cancel()



import os
from unittest.mock import AsyncMock

import pytest

from alist_mikananirss import DownloadManager
from alist_mikananirss.websites.models import ResourceInfo


@pytest.fixture
def base_path():
    return "/base/path"


@pytest.mark.asyncio
async def test_initialize(base_path):
    mock_alist = AsyncMock()
    mock_db = AsyncMock()
    DownloadManager.initialize(mock_alist, base_path, True, True, mock_db)
    assert DownloadManager().base_download_path == base_path


def test_build_download_path_with_anime_name(base_path):
    resource = ResourceInfo(
        resource_title="Test Resource",
        torrent_url="https://example.com/torrent",
        published_date="2023-05-20",
        anime_name="Test Anime",
    )
    expected_path = os.path.join(base_path, "Test Anime")
    test_instance = DownloadManager()
    assert test_instance._build_download_path(resource) == expected_path


def test_build_download_path_with_anime_name_and_season(base_path):
    resource = ResourceInfo(
        resource_title="Test Resource",
        torrent_url="https://example.com/torrent",
        published_date="2023-05-20",
        anime_name="Test Anime",
        season=1,
    )
    expected_path = os.path.join(base_path, "Test Anime", "Season 1")
    test_instance = DownloadManager()
    assert test_instance._build_download_path(resource) == expected_path


def test_build_download_path_with_illegal_characters(base_path):
    # 测试是否能正确处理动画名中的非法字符
    resource = ResourceInfo(
        resource_title="Test Resource",
        torrent_url="https://example.com/torrent",
        published_date="2023-05-20",
        anime_name="Test:Anime!",
        season=2,
    )
    expected_path = os.path.join(base_path, "Test Anime!", "Season 2")
    test_instance = DownloadManager()
    assert test_instance._build_download_path(resource) == expected_path


def test_build_download_path_without_anime_name(base_path):
    # 无动画名时，不会创建子文件夹，将视频文件下载到下载目录的根目录下
    resource_none_name = ResourceInfo(
        resource_title="Test Resource",
        torrent_url="https://example.com/torrent",
        published_date="2023-05-20",
        season=4,
    )
    resource_empty_name = ResourceInfo(
        resource_title="Test Resource",
        torrent_url="https://example.com/torrent",
        published_date="2023-05-20",
        anime_name="",
        season=3,
    )
    expected_path = base_path
    test_instance = DownloadManager()
    assert test_instance._build_download_path(resource_none_name) == expected_path
    assert test_instance._build_download_path(resource_empty_name) == expected_path


def test_build_download_path_with_special_season(base_path):
    resource = ResourceInfo(
        resource_title="Test Resource",
        torrent_url="https://example.com/torrent",
        published_date="2023-05-20",
        anime_name="Test Anime",
        season=0,
    )
    expected_path = os.path.join(base_path, "Test Anime", "Season 0")
    test_instance = DownloadManager()
    assert test_instance._build_download_path(resource) == expected_path



import os
import uuid

import pytest
import pytest_asyncio

from alist_mikananirss.common.database import SubscribeDatabase, db_dirpath
from alist_mikananirss.websites.models import LanguageType, ResourceInfo


@pytest_asyncio.fixture
async def test_db():
    # 为每个测试创建唯一的数据库文件名
    unique_db_name = f"test_db_{uuid.uuid4()}.db"
    db = await SubscribeDatabase.create(unique_db_name)
    db_filepath = os.path.join(db_dirpath, unique_db_name)

    yield db

    # 清理工作
    await db.close()
    if os.path.exists(db_filepath):
        try:
            os.remove(db_filepath)
        except PermissionError:
            import time

            time.sleep(0.1)
            os.remove(db_filepath)


@pytest.mark.asyncio
async def test_create_table(test_db):
    cursor = await test_db.db.execute(
        "SELECT name FROM sqlite_master WHERE type='table';"
    )
    tables = await cursor.fetchall()
    assert ("resource_data",) in tables
    assert ("db_version",) in tables
    await test_db.close()


@pytest.mark.asyncio
async def test_insert_and_check_existence(test_db):
    resource = ResourceInfo(
        resource_title="Test Anime",
        torrent_url="https://example.com/test.torrent",
        published_date="2023-05-20T12:00:00",
        anime_name="Test Anime",
        season=1,
        episode=1,
        fansub="TestSub",
        quality="1080p",
        languages=[LanguageType.SIMPLIFIED_CHINESE],
    )

    await test_db.insert_resource_info(resource)

    assert await test_db.is_resource_title_exist("Test Anime")


@pytest.mark.asyncio
async def test_delete_by_id(test_db):
    await test_db.insert(
        "Test Delete",
        "https://example.com/delete.torrent",
        "2023-05-20T12:00:00",
        "2023-05-20T12:01:00",
        "Test Anime",
    )

    await test_db.connect()
    cursor = await test_db.db.execute(
        "SELECT id FROM resource_data WHERE resource_title=?", ("Test Delete",)
    )
    id_to_delete = await cursor.fetchone()
    id_to_delete = id_to_delete[0]
    await test_db.close()

    await test_db.delete_by_id(id_to_delete)

    assert not await test_db.is_resource_title_exist("Test Delete")


@pytest.mark.asyncio
async def test_delete_by_torrent_url(test_db):
    url = "https://example.com/delete_url.torrent"
    await test_db.insert(
        "Test Delete URL",
        url,
        "2023-05-20T12:00:00",
        "2023-05-20T12:01:00",
        "Test Anime",
    )

    await test_db.delete_by_torrent_url(url)

    assert not await test_db.is_resource_title_exist("Test Delete URL")


@pytest.mark.asyncio
async def test_delete_by_resource_title(test_db):
    title = "Test Delete Title"
    await test_db.insert(
        title,
        "https://example.com/delete_title.torrent",
        "2023-05-20T12:00:00",
        "2023-05-20T12:01:00",
        "Test Anime",
    )

    await test_db.delete_by_resource_title(title)

    assert not await test_db.is_resource_title_exist(title)


@pytest.mark.asyncio
async def test_upgrade_database(test_db):
    await test_db.connect()
    await test_db.db.execute("DROP TABLE IF EXISTS resource_data")
    await test_db.db.execute("DROP TABLE IF EXISTS db_version")
    await test_db.db.execute(
        """
        CREATE TABLE resource_data (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            title TEXT NOT NULL,
            link TEXT UNIQUE,
            published_date TEXT,
            downloaded_date TEXT,
            anime_name TEXT
        )
    """
    )
    await test_db.db.commit()
    await test_db._upgrade_database()
    cursor = await test_db.db.execute("SELECT version FROM db_version")
    version = await cursor.fetchone()
    version = version[0]
    assert version == 1

    cursor = await test_db.db.execute("PRAGMA table_info(resource_data)")
    columns = [info[1] for info in await cursor.fetchall()]
    expected_columns = [
        "id",
        "resource_title",
        "torrent_url",
        "published_date",
        "downloaded_date",
        "anime_name",
        "season",
        "episode",
        "fansub",
        "quality",
        "language",
    ]
    assert all(column in columns for column in expected_columns)
    await test_db.close()


@pytest.mark.asyncio
async def test_insert_duplicate(test_db):
    await test_db.insert(
        "Duplicate Test",
        "https://example.com/duplicate.torrent",
        "2023-05-20T12:00:00",
        "2023-05-20T12:01:00",
        "Test Anime",
    )

    await test_db.insert(
        "Duplicate Test",
        "https://example.com/duplicate.torrent",
        "2023-05-20T12:00:00",
        "2023-05-20T12:01:00",
        "Test Anime",
    )

    await test_db.connect()
    cursor = await test_db.db.execute(
        "SELECT COUNT(*) FROM resource_data WHERE resource_title=?", ("Duplicate Test",)
    )
    count = await cursor.fetchone()
    count = count[0]
    await test_db.close()
    assert count == 1



from enum import StrEnum
from typing import List, Optional

from pydantic import BaseModel, Field

from alist_mikananirss.websites.models import LanguageType, VideoQuality


class ProviderType(StrEnum):
    OPENAI = "openai"
    DEEPSEEK = "deepseek"


class TMDBSearchParam(BaseModel):
    query: str = Field(..., description="Keywords of the anime name to search")


class TMDBTvInfo(BaseModel):
    anime_name: str = Field(..., description="The name of the anime in tmdb")
    tvid: int = Field(..., description="The tmdb id of the anime")


class ResourceTitleExtractResult(BaseModel):
    anime_name: str = Field(..., description="The name of the anime")
    season: int = Field(
        ...,
        description="The season of the anime.Default to be 1. But if special episode, it should be 0",
    )
    episode: int = Field(
        ...,
        description="The episode number. It should be int. If float, it means special episode",
    )
    quality: Optional[VideoQuality] = Field(..., description="The quality of the video")
    fansub: Optional[str] = Field(..., description="The fansub of the video")
    languages: List[LanguageType] = Field(
        ..., description="The subtitle language of the video"
    )
    version: int = Field(
        ..., description="The version of the video's subtitle, default to be 1"
    )


class AnimeNameExtractResult(BaseModel):
    anime_name: str = Field(
        ..., description="The pure name of the anime without season or other info"
    )
    season: int = Field(..., description="The season of the anime")



import os
from datetime import datetime

import aiosqlite
from loguru import logger

from alist_mikananirss.websites.models import ResourceInfo

db_dirpath = "data"
os.makedirs(db_dirpath, exist_ok=True)


class SubscribeDatabase:
    def __init__(self):
        raise RuntimeError("Use `SubscribeDatabase.create()` to create an instance")

    @classmethod
    async def create(cls, db_name="subscribe_database.db"):
        self = cls.__new__(cls)
        self.db_filepath = os.path.join(db_dirpath, db_name)
        self.db = None
        await self.initialize()
        return self

    async def initialize(self):
        if not os.path.exists(self.db_filepath):
            await self.connect()
            await self.__create_table()
        else:
            await self.connect()
            await self._upgrade_database()

    async def connect(self):
        if self.db:
            return
        self.db = await aiosqlite.connect(self.db_filepath)
        if not self.db:
            raise RuntimeError("Failed to connect to database")

    async def close(self):
        if self.db:
            await self.db.close()
            self.db = None

    async def __create_table(self):
        await self.db.execute(
            """
            CREATE TABLE IF NOT EXISTS resource_data (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                resource_title TEXT NOT NULL,
                torrent_url TEXT UNIQUE,
                published_date TEXT,
                downloaded_date TEXT,
                anime_name TEXT,
                season INTEGER,
                episode INTEGER,
                fansub TEXT,
                quality TEXT,
                language TEXT
            )
            """
        )
        await self.db.execute(
            """
            CREATE TABLE IF NOT EXISTS db_version (
                version INTEGER PRIMARY KEY
            )
            """
        )
        await self.db.execute("INSERT INTO db_version (version) VALUES (1)")
        await self.db.commit()

    async def _upgrade_database(self):
        try:
            cursor = await self.db.execute("SELECT version FROM db_version")
            version = await cursor.fetchone()
            version = version[0]
        except aiosqlite.OperationalError:
            await self.db.execute(
                """
                    CREATE TABLE IF NOT EXISTS db_version (
                        version INTEGER PRIMARY KEY
                    )
                """
            )
            version = 0

        if version < 1:
            try:
                await self.db.execute(
                    """
                        CREATE TABLE resource_data_new (
                            id INTEGER PRIMARY KEY AUTOINCREMENT,
                            resource_title TEXT NOT NULL,
                            torrent_url TEXT UNIQUE,
                            published_date TEXT,
                            downloaded_date TEXT,
                            anime_name TEXT,
                            season INTEGER,
                            episode INTEGER,
                            fansub TEXT,
                            quality TEXT,
                            language TEXT
                        )
                    """
                )
                # 迁移数据
                await self.db.execute(
                    """
                        INSERT INTO resource_data_new (resource_title, torrent_url, published_date, downloaded_date, anime_name)
                        SELECT title, link, published_date, downloaded_date, anime_name FROM resource_data
                    """
                )
                # 删除旧表
                await self.db.execute("DROP TABLE resource_data")
                # 重命名新表
                await self.db.execute(
                    "ALTER TABLE resource_data_new RENAME TO resource_data"
                )
                # 创建新的索引
                await self.db.execute(
                    """
                        CREATE UNIQUE INDEX idx_title
                        ON resource_data(resource_title)
                    """
                )
                await self.db.execute(
                    "INSERT OR REPLACE INTO db_version (version) VALUES (1)"
                )
                await self.db.commit()
                logger.info("Database upgraded to version 1")
            except Exception as e:
                logger.error(f"Error during database upgrade: {e}")
                await self.db.rollback()

    async def insert(
        self,
        resource_title,
        torrent_url,
        published_date,
        downloaded_date,
        anime_name,
        season=None,
        episode=None,
        fansub=None,
        quality=None,
        language=None,
    ):
        language_str = "".join(language) if language else None
        try:
            await self.db.execute(
                """
                    INSERT INTO resource_data 
                    (resource_title, torrent_url, published_date, downloaded_date, anime_name, season, episode, fansub, quality, language)
                    VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
                """,
                (
                    resource_title,
                    torrent_url,
                    published_date,
                    downloaded_date,
                    anime_name,
                    season,
                    episode,
                    fansub,
                    quality,
                    language_str,
                ),
            )
            await self.db.commit()
            logger.debug(f"Insert new resource: {anime_name}, {resource_title}")
        except aiosqlite.IntegrityError:
            logger.debug(f"Resource already exists: {anime_name}, {resource_title}")
        except Exception as e:
            logger.error(f"Error when inserting resource: {e}")

    async def insert_resource_info(self, resource: ResourceInfo):
        downloaded_date = datetime.now().strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3]
        await self.insert(
            resource.resource_title,
            resource.torrent_url,
            resource.published_date,
            downloaded_date,
            resource.anime_name,
            season=resource.season,
            episode=resource.episode,
            fansub=resource.fansub,
            quality=resource.quality,
            language=resource.languages,
        )

    async def is_resource_title_exist(self, resource_title: str):
        try:
            cursor = await self.db.execute(
                "SELECT 1 FROM resource_data WHERE resource_title = ? LIMIT 1",
                (resource_title,),
            )
            return await cursor.fetchone() is not None
        except Exception as e:
            logger.error(f"Error checking resource existence: {e}")
            return False

    async def delete_by_id(self, id):
        try:
            await self.db.execute("DELETE FROM resource_data WHERE id=?", (id,))
            await self.db.commit()
            logger.debug(f"Delete resource data: {id}")
        except Exception as e:
            logger.error(f"Error when delete resource data:\n {e}")

    async def delete_by_torrent_url(self, url: str):
        try:
            await self.db.execute(
                "DELETE FROM resource_data WHERE torrent_url=?", (url,)
            )
            await self.db.commit()
            logger.debug(f"Delete resource data: {url}")
        except Exception as e:
            logger.error(f"Error when delete resource data:\n {e}")

    async def delete_by_resource_title(self, title: str):
        try:
            await self.db.execute(
                "DELETE FROM resource_data WHERE resource_title=?", (title,)
            )
            await self.db.commit()
            logger.debug(f"Delete resource data: {title}")
        except Exception as e:
            logger.error(f"Error when delete resource data:\n {e}")



import asyncio
from typing import List

from loguru import logger
from tenacity import retry, stop_after_attempt, wait_exponential

from alist_mikananirss.bot import NotificationBot, NotificationMsg
from alist_mikananirss.websites.models import ResourceInfo

from ..utils import Singleton


class NotificationSender(metaclass=Singleton):

    def __init__(self, notification_bots: list[NotificationBot], interval: int = 60):
        self.notification_bots = notification_bots
        self._interval = interval
        self._queue = asyncio.Queue()
        self._max_retries = 3

    @classmethod
    def initialize(cls, notification_bots: List[NotificationBot], interval: int = 60):
        cls(notification_bots, interval)

    @classmethod
    def set_notification_bots(cls, notification_bots: List[NotificationBot]):
        instance = cls()
        instance.notification_bots = notification_bots

    @classmethod
    async def add_resource(cls, resource: ResourceInfo):
        instance = cls()
        await instance._queue.put(resource)

    @classmethod
    def set_interval(cls, interval: int):
        instance = cls()
        instance._interval = interval

    async def _run(self):
        while True:
            await asyncio.sleep(self._interval)
            resources = []
            while not self._queue.empty():
                try:
                    resource = self._queue.get_nowait()
                    resources.append(resource)
                except asyncio.QueueEmpty:
                    break
            if resources:
                await self._send(resources)

    @classmethod
    async def run(cls):
        instance = cls()
        await instance._run()

    async def _send(self, resources: List[ResourceInfo]):
        if not self.notification_bots:
            return
        msg = NotificationMsg.from_resources(resources)
        logger.debug(f"Send notification\n: {msg}")

        tasks = [self._send_with_retry(bot, msg) for bot in self.notification_bots]
        results = await asyncio.gather(*tasks, return_exceptions=True)

        for result in results:
            if isinstance(result, Exception):
                logger.error(f"Failed to send notification after all retries: {result}")

    @retry(
        stop=stop_after_attempt(3),
        wait=wait_exponential(multiplier=3, min=5, max=30),
        reraise=True,
    )
    async def _send_with_retry(self, bot: NotificationBot, msg: NotificationMsg):
        try:
            await bot.send_message(msg)
        except Exception as e:
            logger.warning(f"Attempt failed for {type(bot.bot)}: {e}")
            raise



import random
from datetime import datetime
from unittest.mock import AsyncMock, patch

import pytest

from alist_mikananirss.alist.api import Alist
from alist_mikananirss.alist.tasks import (
    AlistDeletePolicy,
    AlistDownloaderType,
    AlistDownloadTask,
    AlistTaskState,
    AlistTaskType,
    AlistTransferTask,
)


@pytest.fixture
def alist():
    return Alist(
        base_url="https://example.com",
        token="test_token",
        downloader=AlistDownloaderType.ARIA,
    )


@pytest.fixture
def create_task_json():
    def _create_download_task_json(tid, state, url, download_path, status=None):
        task_json_dict = {
            "creator": "alist_mikananirss",
            "creator_role": 2,
            "end_time": None,
            "error": "",
            "id": tid,
            "name": f"download {url} to ({download_path})",
            "progress": 100,
            "start_time": datetime.now().strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
            "state": state,
            "status": status if status else "[qBittorrent]: [qBittorrent] downloading",
            "total_bytes": random.randint(1000000, 1000000000),
        }
        return task_json_dict

    def _create_transfer_task_json(
        tid, state, uuid, target_drive, target_dir, filename, total_bytes=None
    ):
        task_json_dict = {
            "creator": "twosix",
            "creator_role": 2,
            "end_time": None,
            "error": "",
            "id": tid,
            "name": f"transfer [](/opt/alist/data/temp/qBittorrent/{uuid}/{filename}) to [{target_drive}]({target_dir})",
            "progress": 100,
            "start_time": datetime.now().strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
            "state": state,
            "status": "getting src object",
            "total_bytes": (
                total_bytes if total_bytes else random.randint(1000000, 1000000000)
            ),
        }
        return task_json_dict

    return _create_download_task_json, _create_transfer_task_json


@pytest.mark.asyncio
async def test_alist_init(alist):
    assert alist.base_url == "https://example.com"
    assert alist.token == "test_token"
    assert alist.downloader == AlistDownloaderType.ARIA


@pytest.mark.asyncio
async def test_get_alist_ver(alist):
    with patch.object(alist, "_api_call", new_callable=AsyncMock) as mock_api_call:
        mock_api_call.return_value = {"version": "v2.0.0"}
        version = await alist.get_alist_ver()
        assert version == "2.0.0"
        mock_api_call.assert_called_once_with("GET", "/api/public/settings")


@pytest.mark.asyncio
async def test_add_offline_download_task(alist, create_task_json):
    with patch.object(alist, "_api_call", new_callable=AsyncMock) as mock_api_call:
        create_dl_task_json, _ = create_task_json
        json_data = create_dl_task_json(
            "dl1",
            AlistTaskState.Running,
            "magnet:?xt=xxx",
            "/Local/Anime/Season 1",
        )
        mock_api_call.return_value = {
            "tasks": [json_data],
        }

        tasks = await alist.add_offline_download_task(
            "/Local/Anime/Season 1", ["magnet:?xt=xxx"]
        )
        assert isinstance(tasks[0], AlistDownloadTask)
        mock_api_call.assert_called_once_with(
            "POST",
            "api/fs/add_offline_download",
            json={
                "delete_policy": AlistDeletePolicy.DeleteAlways.value,
                "path": "/Local/Anime/Season 1",
                "urls": ["magnet:?xt=xxx"],
                "tool": alist.downloader.value,
            },
        )


@pytest.mark.asyncio
async def test_list_dir(alist):
    with patch.object(alist, "_api_call", new_callable=AsyncMock) as mock_api_call:
        mock_api_call.return_value = {
            "content": [{"name": "file1.txt"}, {"name": "file2.txt"}]
        }
        files = await alist.list_dir("/test/path")
        assert files == ["file1.txt", "file2.txt"]


@pytest.mark.asyncio
async def test_cancel_task(alist, create_task_json):
    with patch.object(alist, "_api_call", new_callable=AsyncMock) as mock_api_call:
        create_dl_task_json, _ = create_task_json
        json_data = create_dl_task_json(
            "dl1", AlistTaskState.Running, "magnet:?xt=xxx", "/Local/Anime/Season 1"
        )
        task = AlistDownloadTask.from_json(json_data)
        result = await alist.cancel_task(task)
        assert result is True
        mock_api_call.assert_called_once_with(
            "POST", f"/api/task/offline_download/cancel?tid={json_data['id']}"
        )


@pytest.mark.asyncio
async def test_rename(alist):
    with patch.object(alist, "_api_call", new_callable=AsyncMock) as mock_api_call:
        await alist.rename("/old/path/file.txt", "new_file.txt")
        mock_api_call.assert_called_once_with(
            "POST",
            "api/fs/rename",
            json={"path": "/old/path/file.txt", "name": "new_file.txt"},
        )


def test_dl_task_extract(create_task_json):
    create_dl_task_json, _ = create_task_json
    test_url = "magnet:?xt=testtorrent1"
    test_path = "/Local/Anime/Season 1"
    json_data = create_dl_task_json(
        "dl1",
        AlistTaskState.Running,
        test_url,
        test_path,
    )

    task: AlistDownloadTask = AlistDownloadTask.from_json(json_data)

    assert task.url == test_url
    assert task.download_path == test_path
    assert task.task_type == AlistTaskType.DOWNLOAD

    # seeding situation
    json_data = create_dl_task_json(
        "dl1",
        AlistTaskState.Running,
        test_url,
        test_path,
        "offline download completed, waiting for seeding",
    )
    task2 = AlistDownloadTask.from_json(json_data)
    assert task2.state == AlistTaskState.Succeeded


def test_tf_task_extract(create_task_json):
    _, create_tf_task_json = create_task_json
    test_uuid = "d82ddec1-08f6-4894-b7ed-d9c9f25dc4db"
    test_target_drive = "/Google"
    test_target_dir = "/Debug/test/Season 1"
    test_filename = "test.mkv"
    json_data = create_tf_task_json(
        "tf1",
        AlistTaskState.Running,
        test_uuid,
        test_target_drive,
        test_target_dir,
        test_filename,
    )
    task: AlistTransferTask = AlistTransferTask.from_json(
        json_data,
    )

    assert task.uuid == test_uuid
    assert task.target_path == test_target_drive + test_target_dir + "/" + test_filename
    assert task.task_type == AlistTaskType.TRANSFER

    # openlist case
    json_data = create_tf_task_json(
        "tf1",
        AlistTaskState.Running,
        test_uuid,
        test_target_drive,
        test_target_dir,
        "test.mp4",
    )
    json_data["name"] = (
        "transfer [](/data/tmp/qBittorrent/f32e8969-8a6c-4878-8932-ca3318f9933e/Summer Pockets - S01E14 - [三明治摆烂组][简体内嵌][H264 8bit 1080P].mp4) to [/crypt-gd1](/)"
    )
    task2: AlistTransferTask = AlistTransferTask.from_json(json_data)
    assert task2.uuid == "f32e8969-8a6c-4878-8932-ca3318f9933e"
    assert (
        task2.target_path
        == "/crypt-gd1/Summer Pockets - S01E14 - [三明治摆烂组][简体内嵌][H264 8bit 1080P].mp4"
    )
    assert task2.task_type == AlistTaskType.TRANSFER
